//
// WTFeedbackView.m    1.0
//
// Created by Wagner Truppel on 2009-10-08.
// Copyright 2009 Wagner Truppel (jrcapab@gmail.com). All rights reserved.
//
// Inspired by and modified from DSActivityView.
//
// =======================================================
// DSActivityView
// Dejal Open Source
//
// Created by David Sinclair on 2009-07-26.
// Copyright 2009 Dejal Systems, LLC. All rights reserved.
// =======================================================
//
// Redistribution and use in binary and source forms, with or without
// modification, are permitted for any project, commercial or otherwise,
// provided that the following conditions are met:
//
// Redistributions in binary form must display the copyright notice in the
// About view, website, and/or documentation.
//
// Redistributions of source code must retain the copyright notice, this
// list of conditions, and the following disclaimer.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THIS SOFTWARE.
//
// Credit: inspired by Matt Gallagher's LoadingView blog post:
// http://cocoawithlove.com/2009/04/showing-message-over-iphone-keyboard.html
//

#import <QuartzCore/QuartzCore.h>
#import "WTFeedbackView.h"





// ========================================================================= //
                          #pragma mark string constants
// ========================================================================= //

static NSString* const kWTFeedbackViewDefaultLabelText =
    @"Loading...";

static NSString* const kWTFeedbackViewDefaultLabelTextLocalizationString =
    @"Default WTFeedbackView label text";

// ========================================================================= //





// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //
                                #pragma mark -
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //

@interface WTFeedbackView (Private)

+ (NSString*) defaultLabelText;

- (UIView*) borderView;

- (UILabel*) feedbackLabel;

- (UIActivityIndicatorView*) activityIndicatorView;

- (UIProgressView*) progressView;

- (void) doRemoveView;

- (void) doSetText: (NSString*) text;

- (void) textChangeFrameResizingAnimDidStop: (NSString*) animationID
                                   finished: (NSNumber*) finished
                                    context: (void*) context;

- (void) doSwitchToActivityIndicatorView;

- (void) doSwitchToProgressView;

- (void) doUpdateProgress: (NSNumber*) floatObj;

- (void) hideProgressAnimDidStop: (NSString*) animationID
                        finished: (NSNumber*) finished
                         context: (void*) context;

- (void) hideActivityAnimDidStop: (NSString*) animationID
                        finished: (NSNumber*) finished
                         context: (void*) context;

- (void) toActivityFrameAnimDidStop: (NSString*) animationID
                           finished: (NSNumber*) finished
                            context: (void*) context;

- (void) toProgressFrameAnimDidStop: (NSString*) animationID
                           finished: (NSNumber*) finished
                            context: (void*) context;

- (void) switchAnimDidStop: (NSString*) animationID
                  finished: (NSNumber*) finished
                   context: (void*) context;

- (void) showAnimDidStop: (NSString*) animationID
                finished: (NSNumber*) finished
                 context: (void*) context;

- (void) removeAnimDidStop: (NSString*) animationID
                  finished: (NSNumber*) finished
                   context: (void*) context;

- (void) layoutSubviewsStyleSimple;

- (void) layoutBorderAndActivity;
- (void) layoutBorderAndProgress;

- (void) layoutCenteredLabel;

@end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //
                                #pragma mark -
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //

@implementation WTFeedbackView

// ========================================================================= //
                          #pragma mark properties
// ========================================================================= //

@synthesize feedbackKind                 = _feedbackKind;
@synthesize feedbackStyle                = _feedbackStyle;

@synthesize labelTextSize                = _labelTextSize;
@synthesize textBorderFrame              = _textBorderFrame;

@synthesize originalView                 = _originalView;
@synthesize   borderView                 = _borderView;

@synthesize feedbackLabel                = _feedbackLabel;

@synthesize activityIndicatorView        = _activityIndicatorView;
@synthesize          progressView        = _progressView;

@synthesize currentlyShowingProgressView = _currentlyShowingProgressView;

@synthesize activityBorderFrame          = _activityBorderFrame;
@synthesize progressBorderFrame          = _progressBorderFrame;

// ========================================================================= //
                       #pragma mark static variables
// ========================================================================= //

static WTFeedbackView* wtFeedbackView = nil;

// ========================================================================= //
                         #pragma mark factory methods
// ========================================================================= //

// Creates a simple feedback view, containing an activity indicator and
// the feedback label, horizontally next to one another, with the entire
// feedback view centered within the specified view.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newSimpleActivityViewForView: (UIView*) addToView;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newSimpleActivityViewForView: addToView
                                             withText: labelText];
}

// ========================================================================= //

// Creates a simple feedback view, containing an activity indicator and
// the feedback label, horizontally next to one another, with the entire
// feedback view centered within the specified view.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newSimpleActivityViewForView: (UIView*) addToView
                                        withText: (NSString*) labelText;
{
    // Not autoreleased because the feedback view is essentially a singleton.

    return [[[self class] alloc] initWithKind: WTFeedbackViewKindActivity
                                        style: WTFeedbackViewStyleSimple
                                      forView: addToView
                                     withText: labelText
                          showingProgressView: NO];
}

// ========================================================================= //

// Creates a bezel-style feedback view, containing an activity indicator and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the specified view.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newBezelActivityViewForView: (UIView*) addToView;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newBezelActivityViewForView: addToView
                                            withText: labelText];
}

// ========================================================================= //

// Creates a bezel-style feedback view, containing an activity indicator and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the specified view.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newBezelActivityViewForView: (UIView*) addToView
                                       withText: (NSString*) labelText;
{
    // Not autoreleased because the feedback view is essentially a singleton.

    return [[[self class] alloc] initWithKind: WTFeedbackViewKindActivity
                                        style: WTFeedbackViewStyleBezel
                                      forView: addToView
                                     withText: labelText
                          showingProgressView: NO];
}

// ========================================================================= //

// Creates a bezel-style feedback view, containing a progress view and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the specified view.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newBezelProgressViewForView: (UIView*) addToView;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newBezelProgressViewForView: addToView
                                            withText: labelText];
}

// ========================================================================= //

// Creates a bezel-style feedback view, containing a progress view and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the specified view.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newBezelProgressViewForView: (UIView*) addToView
                                       withText: (NSString*) labelText;
{
    // Not autoreleased because the feedback view is essentially a singleton.

    return [[[self class] alloc] initWithKind: WTFeedbackViewKindProgress
                                        style: WTFeedbackViewStyleBezel
                                      forView: addToView
                                     withText: labelText
                          showingProgressView: YES];
}

// ========================================================================= //

// Creates a bezel-style feedback view, containing boh an activity indicator
// and a progress view (though only one can be shown at any given time), and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the specified view. Which indicator (activity
// or progress) is visible when the feedback view is first displayed is set
// by the specified boolean. The visible indicator can be switched on demand
// at any time, while the feedback view is showing.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newBezelFlexibleViewForView: (UIView*) addToView
                            showingProgressView: (BOOL) showing;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newBezelFlexibleViewForView: addToView
                                 showingProgressView: showing
                                            withText: labelText];
}

// ========================================================================= //

// Creates a bezel-style feedback view, containing boh an activity indicator
// and a progress view (though only one can be shown at any given time), and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the specified view. Which indicator (activity
// or progress) is visible when the feedback view is first displayed is set
// by the specified boolean. The visible indicator can be switched on demand
// at any time, while the feedback view is showing.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the specified
// view.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newBezelFlexibleViewForView: (UIView*) addToView
                            showingProgressView: (BOOL) showing
                                       withText: (NSString*) labelText;
{
    // Not autoreleased because the feedback view is
    // essentially a singleton.

    return [[[self class] alloc] initWithKind: WTFeedbackViewKindFlexible
                                        style: WTFeedbackViewStyleBezel
                                      forView: addToView
                                     withText: labelText
                          showingProgressView: showing];
}

// ========================================================================= //

// Creates a keyboard-style feedback view, containing an activity
// indicator and the feedback label, vertically next to one another,
// with the entire feedback view centered within the keyboard view,
// if the keyboard is showing.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the keyboard
// view, or nil if the keyboard isn't showing.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newKeyboardActivityView;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newKeyboardActivityViewWithText: labelText];
}

// ========================================================================= //

// Creates a keyboard-style feedback view, containing an activity
// indicator and the feedback label, vertically next to one another,
// with the entire feedback view centered within the keyboard view,
// if the keyboard is showing.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the keyboard
// view, or nil if the keyboard isn't showing.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newKeyboardActivityViewWithText: (NSString*) labelText;
{
    UIView* keyboardView = [[UIApplication sharedApplication] keyboardView];

    if (keyboardView == nil)
    { return nil; }
    else
    {
        // Not autoreleased because the feedback view is essentially
        // a singleton.

        return [[[self class] alloc] initWithKind: WTFeedbackViewKindActivity
                                            style: WTFeedbackViewStyleKeyboard
                                          forView: keyboardView
                                         withText: labelText
                              showingProgressView: NO];
    }
}

// ========================================================================= //

// Creates a keyboard-style feedback view, containing a progress
// view and the feedback label, vertically next to one another,
// with the entire feedback view centered within the keyboard view,
// if the keyboard is showing.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the keyboard
// view, or nil if the keyboard isn't showing.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newKeyboardProgressView;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newKeyboardProgressViewWithText: labelText];
}

// ========================================================================= //

// Creates a keyboard-style feedback view, containing a progress
// view and the feedback label, vertically next to one another,
// with the entire feedback view centered within the keyboard view,
// if the keyboard is showing.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the keyboard
// view, or nil if the keyboard isn't showing.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newKeyboardProgressViewWithText: (NSString*) labelText;
{
    UIView* keyboardView = [[UIApplication sharedApplication] keyboardView];

    if (keyboardView == nil)
    { return nil; }
    else
    {
        // Not autoreleased because the feedback view is essentially
        // a singleton.

        return [[[self class] alloc] initWithKind: WTFeedbackViewKindProgress
                                            style: WTFeedbackViewStyleKeyboard
                                          forView: keyboardView
                                         withText: labelText
                              showingProgressView: YES];
    }
}

// ========================================================================= //

// Creates a keyboard-style feedback view, containing boh an activity indicator
// and a progress view (though only one can be shown at any given time), and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the keyboard view, if the keyboard is showing.
// Which indicator (activity or progress) is visible when the feedback view is
// first displayed is set by the specified boolean. The visible indicator can
// be switched on demand at any time, while the feedback view is showing.

// The label text is set to a configurable default value, though it
// can be changed on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the keyboard
// view, or nil if the keyboard isn't showing.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newKeyboardFlexibleViewShowingProgressView: (BOOL) showing;
{
    NSString* labelText = [self defaultLabelText];

    return [[self class] newKeyboardFlexibleViewShowingProgressView: showing
                                                           withText: labelText];
}

// ========================================================================= //

// Creates a keyboard-style feedback view, containing boh an activity indicator
// and a progress view (though only one can be shown at any given time), and
// the feedback label, vertically next to one another, with the entire
// feedback view centered within the keyboard view, if the keyboard is showing.
// Which indicator (activity or progress) is visible when the feedback view is
// first displayed is set by the specified boolean. The visible indicator can
// be switched on demand at any time, while the feedback view is showing.

// The label text is set to the specified string, though it can be changed
// on demand at any time while the feedback view is visible.

// Returns the feedback view, already added as a subview of the keyboard
// view, or nil if the keyboard isn't showing.

// NOTE: this method returns a retained object. However, there is no need
// for the client code to release it, as that is done when one of the
// dismissal methods is invoked.

+ (WTFeedbackView*) newKeyboardFlexibleViewShowingProgressView: (BOOL) showing
                                               withText: (NSString*) labelText;
{
    UIView* keyboardView = [[UIApplication sharedApplication] keyboardView];

    if (keyboardView == nil)
    { return nil; }
    else
    {
        // Not autoreleased because the feedback view is essentially
        // a singleton.

        return [[[self class] alloc] initWithKind: WTFeedbackViewKindFlexible
                                            style: WTFeedbackViewStyleKeyboard
                                          forView: keyboardView
                                         withText: labelText
                              showingProgressView: showing];
    }
}

// ========================================================================= //
                          #pragma mark accessor methods
// ========================================================================= //

// Returns the currently displayed feedback view, or nil if none is showing.

+ (WTFeedbackView*) currentFeedbackView;
{
    return wtFeedbackView;
}

// ========================================================================= //

// Returns the kind of the currently displayed feedback view, or
// WTFeedbackViewKindUnavailable if none is showing.

+ (WTFeedbackViewKind) kind;
{
    if (wtFeedbackView == nil)
    { return WTFeedbackViewKindUnavailable; }
    else
    { return wtFeedbackView.feedbackKind; }
}

// ========================================================================= //

// Returns the style of the currently displayed feedback view, or
// WTFeedbackViewStyleUnavailable if none is showing.

+ (WTFeedbackViewStyle) style;
{
    if (wtFeedbackView == nil)
    { return WTFeedbackViewKindUnavailable; }
    else
    { return wtFeedbackView.feedbackStyle; }
}

// ========================================================================= //

// Returns the default text to be shown.
// The default implementation returns @"Loading...".

+ (NSString*) defaultFeedbackText;
{
    return kWTFeedbackViewDefaultLabelText;
}

// ========================================================================= //

// Returns the default localization string for the feedback text.
// The default implementation returns @"Default WTFeedbackView label text".

+ (NSString*) defaultFeedbackTextLocalizationString;
{
    return kWTFeedbackViewDefaultLabelTextLocalizationString;
}

// ========================================================================= //

// Returns the feedback view's border view, the view that contains the
// activity indicator view and/or the progress view, and the feedback
// label.

// It's accessible so as to allow changing its appearance while the feedback
// view is showing. Any such changes should be executed in the main thread.

// The default implementation creates and returns a transparent view,
// with flexible margins.

// The bezel style has a semi-transparent (alpha = 0.5) black color, with
// a rounded rectangle (corner radius = 10.0).

// The keyboard style has a transparent border.

+ (UIView*) borderView;
{
    return wtFeedbackView.borderView;
}

// ========================================================================= //

// Returns the label displaying the feedback text.

// It's accessible so as to allow changing its appearance while the feedback
// view is showing. Any such changes should be executed in the main thread.

// The label's actual text should be changed using the class method +setText:,
// rather than directly, since +setText: guarantees thread-safety and also
// automatically repositions the subviews of the feedback view, as needed.

// The simple style has the following properties:

// backgroundColor  = [UIColor clearColor];
// font             = [UIFont systemFontOfSize: [UIFont systemFontSize]];
// textAlignment    = UITextAlignmentLeft;
// textColor        = [UIColor blackColor];
// shadowColor      = [UIColor whiteColor];
// shadowOffset     = CGSizeMake(0.0, 1.0);
// autoresizingMask = UIViewAutoresizingFlexibleRightMargin |
//                    UIViewAutoresizingFlexibleTopMargin;

// The bezel and keyboard styles have the following properties:

// backgroundColor  = [UIColor clearColor];
// font             = [UIFont boldSystemFontOfSize: [UIFont systemFontSize]];
// textAlignment    = UITextAlignmentCenter;
// textColor        = [UIColor whiteColor];
// autoresizingMask = UIViewAutoresizingFlexibleLeftMargin  |
//                    UIViewAutoresizingFlexibleRightMargin |
//                    UIViewAutoresizingFlexibleTopMargin;

+ (UILabel*) feedbackLabel;
{
    return wtFeedbackView.feedbackLabel;
}

// ========================================================================= //

// Returns the activity indicator view, or nil for feedback views of kind
// WTFeedbackViewKindProgress.

// It's accessible so as to allow changing its appearance while the feedback
// view is showing. Any such changes should be executed in the main thread.

// The default implementation returns nil.

// The simple style returns a small gray activity indicator.

// The bezel and keyboard styles return nil for the progress
// kind and a large white activity indicator for the activity
// and flexible kinds.

+ (UIActivityIndicatorView*) activityIndicatorView;
{
    return wtFeedbackView.activityIndicatorView;
}

// ========================================================================= //

// Returns the progress view, or nil for feedback views of kind
// WTFeedbackViewKindActivity.

// It's accessible so as to allow changing its appearance while the feedback
// view is showing. Any such changes should be executed in the main thread.

// The default implementation returns nil. Only the
// progress and flexible kinds return a non-nil object.

+ (UIProgressView*) progressView;
{
    return wtFeedbackView.progressView;
}

// ========================================================================= //

// Returns whether or not the feedback view is currently showing its progress
// view, as opposed to its activity indicator view. Always returns NO for a
// feedback view of kind WTFeedbackViewKindActivity and always returns YES for
// a feedback view of kind WTFeedbackViewKindProgress.

// This method is thread-safe.

+ (BOOL) currentlyShowingProgressView;
{
    return wtFeedbackView.currentlyShowingProgressView;
}

// ========================================================================= //
                         #pragma mark mutator methods
// ========================================================================= //

// Sets the label text of the feedback view, automatically resizing and
// recentering the feedback view, if necessary.

// This method is guaranteed to execute its body in the main thread, so
// it's safe to invoke this method from any thread.

// This method should not be overriden.

+ (void) setText: (NSString*) text;
{
    [wtFeedbackView performSelectorOnMainThread: @selector(doSetText:)
                                     withObject: text waitUntilDone: NO];
}

// ========================================================================= //

// For feedback views of kind WTFeedbackViewKindFlexible, hides the progress
// view and shows the activity indicator view. For feedback views of the other
// two kinds, it does nothing.

// This method is guaranteed to execute its body in the main thread, so
// it's safe to invoke this method from any thread.

// This method should not be overriden. Override the associated instance
// method instead, if necessary.

+ (void) switchToActivityIndicatorView;
{
    [wtFeedbackView performSelectorOnMainThread:
            @selector(doSwitchToActivityIndicatorView)
                withObject: nil waitUntilDone: NO];
}

// ========================================================================= //

// For feedback views of kind WTFeedbackViewKindFlexible, hides the activity
// indicator view and shows the progress view. For feedback views of the other
// two kinds, it does nothing.

// This method is guaranteed to execute its body in the main thread, so
// it's safe to invoke this method from any thread.

// This method should not be overriden. Override the associated instance
// method instead, if necessary.

+ (void) switchToProgressView;
{
    [wtFeedbackView performSelectorOnMainThread:
            @selector(doSwitchToProgressView)
                withObject: nil waitUntilDone: NO];
}

// ========================================================================= //

// Does nothing for a feedback view of kind WTFeedbackViewKindActivity. For
// the other two kinds, updates the progress view, even if it's currently
// not showing.

// This method is guaranteed to execute its body in the main thread, so
// it's safe to invoke this method from any thread.

// This method should not be overriden. Override the associated instance
// method instead, if necessary.

+ (void) updateProgress: (CGFloat) progress;
{
    NSNumber* floatObj = [[NSNumber alloc] initWithFloat: progress];
    [wtFeedbackView performSelectorOnMainThread: @selector(doUpdateProgress:)
                                     withObject: floatObj waitUntilDone: NO];
    [floatObj release];
}

// ========================================================================= //

// Immediately removes and releases the feedback view, without any animation.

// This method is guaranteed to execute its body in the main thread, so
// it's safe to invoke this method from any thread.

// This method should not be overriden.

+ (void) removeView;
{
    [wtFeedbackView performSelectorOnMainThread: @selector(doRemoveView)
                                     withObject: nil waitUntilDone: NO];
}

// ========================================================================= //

// Removes and releases the feedback view, with or without an animation.
// The simple style does not support animating the feedback view removal,
// in which case the boolean argument is ignored, and +removeView is
// executed, instead.

// This method is guaranteed to execute its body in the main thread, so
// it's safe to invoke this method from any thread.

// This method should not be overriden. Override the associated instance
// method instead, if necessary.

+ (void) removeViewAnimating: (BOOL) animating;
{
    if (animating)
    {
        [wtFeedbackView performSelectorOnMainThread: @selector(animateRemove)
                                         withObject: nil waitUntilDone: NO];
    }
    else
    {
        [wtFeedbackView performSelectorOnMainThread: @selector(doRemoveView)
                                         withObject: nil waitUntilDone: NO];
    }
}

// ========================================================================= //
                    #pragma mark memory management methods
// ========================================================================= //

// Designated initializer. Configures a feedback view of the specified kind
// and style, using the specified label text, and adds it as a subview of the
// specified view. If the feedback kind is anything other than
// WTFeedbackViewKindFlexible, the boolean argument is ignored (though the
// associated instance variable is correctly set anyway).

- (id) initWithKind: (WTFeedbackViewKind) kind
              style: (WTFeedbackViewStyle) style
            forView: (UIView*) addToView
           withText: (NSString*) labelText
showingProgressView: (BOOL) showing;
{
    self = [super initWithFrame: CGRectZero];

	if (self == nil)
	{ return nil; }

    // Immediately remove any existing feedback view.
    if (wtFeedbackView != nil)
    { [[self class] removeView]; }

    // Remember the new view (it is already retained).
    wtFeedbackView = self;

    self.feedbackKind  = kind;
    self.feedbackStyle = style;

    // Allow subclasses to change the view to which to add the
    // feedback view (e.g. to cover the keyboard).
    self.originalView = addToView;
    addToView = [self viewForView: addToView];

    // Configure this view (the background) and the label text
    // (the label is automatically created).
    [self setupBackground];

    self.feedbackLabel.text = labelText;
    self.labelTextSize = [labelText sizeWithFont: self.feedbackLabel.font];

    // Assemble the subviews (they're automatically created).

	[addToView addSubview: self];
    [self addSubview: self.borderView];

    switch (kind)
    {
        default:
        case WTFeedbackViewKindActivity:
        {
            [self.borderView addSubview: self.activityIndicatorView];
            [self.activityIndicatorView startAnimating];
             self.currentlyShowingProgressView = NO;
        }
        break;

        case WTFeedbackViewKindProgress:
        {
            [self.borderView addSubview: self.progressView];
             self.currentlyShowingProgressView = YES;
        }
        break;

        case WTFeedbackViewKindFlexible:
        {
            [self.borderView addSubview: self.progressView];
            [self.borderView addSubview: self.activityIndicatorView];
             self.currentlyShowingProgressView = showing;

            if (showing)
            { self.activityIndicatorView.alpha = 0.0; }
            else
            {
                 self.progressView.alpha = 0.0;
                [self.activityIndicatorView startAnimating];
            }
        }
        break;
    }

    [self.borderView addSubview: self.feedbackLabel];

    // Animate the view in, if appropriate.
    [self performSelectorOnMainThread: @selector(animateShow)
                           withObject: nil waitUntilDone: NO];

	return self;
}

// ========================================================================= //

- (void) dealloc;
{
    [_originalView  release];
    [_borderView    release];
    [_feedbackLabel release];

    [_activityIndicatorView release];
    [_progressView          release];

    wtFeedbackView = nil;

    [super dealloc];
}

// ========================================================================= //
                        #pragma mark layout methods
// ========================================================================= //

// Positions and sizes the various views that make up the feedback view,
// including after rotation.

- (void) layoutSubviews;
{
    // If we're animating a transform, don't lay out now, since we
    // can't use the frame property when transforming.
    if (! CGAffineTransformIsIdentity(self.borderView.transform))
    { return; }

    self.frame = [self enclosingFrame];

    CGSize textSize = self.labelTextSize;

    // If there's no label text, don't need to allow height for it.
    if (self.feedbackLabel.text.length == 0)
    {
        textSize.height = 0.0;
        self.labelTextSize = textSize;
    }

    if (self.feedbackStyle == WTFeedbackViewStyleSimple)
    { [self layoutSubviewsStyleSimple]; }
    else
    if (self.feedbackKind == WTFeedbackViewKindActivity)
    {
        [self layoutBorderAndActivity];
        [self layoutCenteredLabel];
    }
    else
    if (self.feedbackKind == WTFeedbackViewKindProgress)
    {
        [self layoutBorderAndProgress];
        [self layoutCenteredLabel];
    }
    else
    if (self.feedbackKind == WTFeedbackViewKindFlexible)
    {
        [self layoutBorderAndActivity];
        [self layoutBorderAndProgress];

        if (self.currentlyShowingProgressView)
        { self.borderView.frame = self.progressBorderFrame; }
        else
        { self.borderView.frame = self.activityBorderFrame; }

        [self layoutCenteredLabel];
    }
}

// ========================================================================= //
                       #pragma mark overridable methods
// ========================================================================= //

// Returns the view to which the feedback view is added.

// The default behavior is to return the same view passed as the argument.

// For the bezel style, if there is a keyboard displayed, the view
// is changed to the keyboard's superview.

// For the keyboard style, returns the same view (which should
// already be the keyboard).

- (UIView*) viewForView: (UIView*) view;
{
    if (self.feedbackStyle == WTFeedbackViewStyleBezel)
    {
        UIView* keyboardView =
            [[UIApplication sharedApplication] keyboardView];

        if (keyboardView != nil)
        { view = keyboardView.superview; }
    }

    return view;
}

// ========================================================================= //

// Returns the frame to use for the feedback view.
// The default behavior is to return the superview's bounds.

// For the bezel style, if there is a keyboard displayed, the frame
// is changed to cover the keyboard too.

- (CGRect) enclosingFrame;
{
    CGRect frame = self.superview.bounds;

    if (self.feedbackStyle == WTFeedbackViewStyleBezel)
    {
        if (self.superview != self.originalView)
        {
            frame = [self.originalView convertRect: self.originalView.bounds
                                            toView: self.superview];
        }
    }

    return frame;
}

// ========================================================================= //

// Configures the feedback view's background. The default is to have
// a non-opaque background with an auto-resizing mask equal to
// UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight.

// For the bezel style, the background has a semi-transparent black color
// (alpha = 0.5).

// For the keyboard style, the background has a semi-transparent black
// color that's darker than that for the bezel style (alpha = 0.8).

- (void) setupBackground;
{
	self.opaque = NO;
    self.autoresizingMask = UIViewAutoresizingFlexibleWidth |
                            UIViewAutoresizingFlexibleHeight;

    if (self.feedbackStyle == WTFeedbackViewStyleBezel)
    {
        self.backgroundColor = [[UIColor blackColor]
            colorWithAlphaComponent: 0.5];
    }
    else
    if (self.feedbackStyle == WTFeedbackViewStyleKeyboard)
    {
        self.backgroundColor = [[UIColor blackColor]
            colorWithAlphaComponent: 0.8];
    }
}

// ========================================================================= //

// Animates the feedback view into visibility.
// The default behavior is to do nothing.

// For the bezel style, fades in the background and zooms the
// bezel down from a large size.

// For the keyboard style, simply fades the feedback view in.

// This method is guaranteed to execute in the main thread.

- (void) animateShow;
{
    if (self.feedbackStyle == WTFeedbackViewStyleBezel)
    {
        self.alpha = 0.0;
        self.borderView.transform = CGAffineTransformMakeScale(3.0, 3.0);

        [UIView beginAnimations: nil context: nil];
// [UIView setAnimationDuration: 5.0]; // Animate in slow motion.
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(showAnimDidStop: finished: context:)];

            self.borderView.transform = CGAffineTransformIdentity;
            self.alpha = 1.0;

        [UIView commitAnimations];
    }
    else
    if (self.feedbackStyle == WTFeedbackViewStyleKeyboard)
    {
        self.alpha = 0.0;

        [UIView beginAnimations: nil context: nil];
        [UIView setAnimationDuration: 0.5];
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(showAnimDidStop: finished: context:)];

            self.alpha = 1.0;

        [UIView commitAnimations];
    }
}

// ========================================================================= //

// Animates the feedback view out of visibiltiy.
// The default behavior is to do nothing.

// For the bezel style, fades out the background and zooms the
// bezel down to half size.

// For the keyboard style, simply fades the feedback view out.

// This method is guaranteed to execute in the main thread.

- (void) animateRemove;
{
    if (self.feedbackStyle == WTFeedbackViewStyleBezel)
    {
        self.borderView.transform = CGAffineTransformIdentity;

        [UIView beginAnimations: nil context: nil];
// [UIView setAnimationDuration: 5.0]; // Animate in slow motion.
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(removeAnimDidStop: finished: context:)];

            self.borderView.transform = CGAffineTransformMakeScale(0.5, 0.5);
            self.alpha = 0.0;

        [UIView commitAnimations];
    }
    else
    if (self.feedbackStyle == WTFeedbackViewStyleKeyboard)
    {
        [UIView beginAnimations: nil context: nil];
        [UIView setAnimationDuration: 0.5];
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(removeAnimDidStop: finished: context:)];

            self.alpha = 0.0;

        [UIView commitAnimations];
    }
}

// ========================================================================= //
                        #pragma mark private methods
// ========================================================================= //

+ (NSString*) defaultLabelText;
{
    return NSLocalizedString([[self class] defaultFeedbackText],
        [[self class] defaultFeedbackTextLocalizationString]);
}

// ========================================================================= //

// Returns the feedback view's border view, the view that contains the
// activity indicator view and/or the progress view, and the feedback
// label, creating if if necessary.

// The default implementation creates and returns a transparent view,
// with flexible margins.

// The bezel style has a semi-transparent (alpha = 0.5) black color, with
// a rounded rectangle (corner radius = 10.0).

// The keyboard style has a transparent border.

- (UIView*) borderView;
{
    if (_borderView == nil)
    {
        _borderView = [[UIView alloc] initWithFrame: CGRectZero];

        _borderView.opaque = NO;
        _borderView.backgroundColor = nil;
        _borderView.autoresizingMask = UIViewAutoresizingFlexibleLeftMargin  |
                                       UIViewAutoresizingFlexibleRightMargin |
                                       UIViewAutoresizingFlexibleTopMargin   |
                                       UIViewAutoresizingFlexibleBottomMargin;

        if (self.feedbackStyle == WTFeedbackViewStyleBezel)
        {
            _borderView.backgroundColor = [[UIColor blackColor]
                colorWithAlphaComponent: 0.5];

            _borderView.layer.cornerRadius = 10.0;
        }
    }

    return _borderView;
}

// ========================================================================= //

// Returns the label displaying the feedback text, creating it if necessary.

// The simple style has the following properties:

// backgroundColor  = [UIColor clearColor];
// font             = [UIFont systemFontOfSize: [UIFont systemFontSize]];
// textAlignment    = UITextAlignmentLeft;
// textColor        = [UIColor blackColor];
// shadowColor      = [UIColor whiteColor];
// shadowOffset     = CGSizeMake(0.0, 1.0);
// autoresizingMask = UIViewAutoresizingFlexibleRightMargin |
//                    UIViewAutoresizingFlexibleTopMargin |
//                    UIViewAutoresizingFlexibleBottomMargin;

// The bezel and keyboard styles have the following properties:

// backgroundColor  = [UIColor clearColor];
// font             = [UIFont boldSystemFontOfSize: [UIFont systemFontSize]];
// textAlignment    = UITextAlignmentCenter;
// textColor        = [UIColor whiteColor];
// autoresizingMask = UIViewAutoresizingFlexibleLeftMargin  |
//                    UIViewAutoresizingFlexibleRightMargin |
//                    UIViewAutoresizingFlexibleTopMargin;

- (UILabel*) feedbackLabel;
{
    if (_feedbackLabel == nil)
    {
        _feedbackLabel = [[UILabel alloc] initWithFrame: CGRectZero];
        _feedbackLabel.backgroundColor = [UIColor clearColor];

        if (self.feedbackStyle == WTFeedbackViewStyleSimple)
        {
            _feedbackLabel.font = [UIFont systemFontOfSize:
                [UIFont systemFontSize]];

            _feedbackLabel.textAlignment = UITextAlignmentLeft;
            _feedbackLabel.textColor     = [UIColor blackColor];
            _feedbackLabel.shadowColor   = [UIColor whiteColor];
            _feedbackLabel.shadowOffset  = CGSizeMake(0.0, 1.0);

            _feedbackLabel.autoresizingMask =
                UIViewAutoresizingFlexibleRightMargin |
                UIViewAutoresizingFlexibleTopMargin |
                UIViewAutoresizingFlexibleBottomMargin;
        }
        else
        {
            _feedbackLabel.font = [UIFont boldSystemFontOfSize:
                [UIFont systemFontSize]];

            _feedbackLabel.textAlignment = UITextAlignmentCenter;
            _feedbackLabel.textColor     = [UIColor whiteColor];

            _feedbackLabel.autoresizingMask =
                UIViewAutoresizingFlexibleLeftMargin  |
                UIViewAutoresizingFlexibleRightMargin |
                UIViewAutoresizingFlexibleTopMargin;
        }
    }

    return _feedbackLabel;
}

// ========================================================================= //

// Returns the activity indicator view.

// The default implementation returns nil.

// The simple style returns a small gray activity indicator.

// The bezel and keyboard styles return nil for the progress
// kind and a large white activity indicator for the activity
// and flexible kinds.

- (UIActivityIndicatorView*) activityIndicatorView;
{
    if (self.feedbackStyle == WTFeedbackViewStyleSimple)
    {
        if (_activityIndicatorView == nil)
        {
            _activityIndicatorView = [[UIActivityIndicatorView alloc]
                initWithActivityIndicatorStyle:
                    UIActivityIndicatorViewStyleGray];

            _activityIndicatorView.autoresizingMask =
                UIViewAutoresizingFlexibleRightMargin |
                UIViewAutoresizingFlexibleTopMargin |
                UIViewAutoresizingFlexibleBottomMargin;
        }

        return _activityIndicatorView;
    }
    else
    if (self.feedbackKind != WTFeedbackViewKindProgress)
    {
        if (_activityIndicatorView == nil)
        {
            _activityIndicatorView = [[UIActivityIndicatorView alloc]
                initWithActivityIndicatorStyle:
                    UIActivityIndicatorViewStyleWhiteLarge];

            _activityIndicatorView.autoresizingMask =
                UIViewAutoresizingFlexibleLeftMargin  |
                UIViewAutoresizingFlexibleRightMargin |
                UIViewAutoresizingFlexibleTopMargin;
        }

        return _activityIndicatorView;
    }

    return nil;
}

// ========================================================================= //

// Returns the progress view.

// The default implementation returns nil. Only the
// progress and flexible kinds return a non-nil object.

- (UIProgressView*) progressView;
{
    if (self.feedbackKind != WTFeedbackViewKindActivity)
    {
        if (_progressView == nil)
        {
            _progressView = [[UIProgressView alloc]
                initWithProgressViewStyle: UIProgressViewStyleDefault];
    //          initWithProgressViewStyle: UIProgressViewStyleBar];

            _progressView.autoresizingMask =
                UIViewAutoresizingFlexibleWidth |
                UIViewAutoresizingFlexibleTopMargin;
        }

        return _progressView;
    }

    return nil;
}

// ========================================================================= //

// Immediately removes the feedback view, without any animation.

- (void) doRemoveView;
{
    [self removeFromSuperview];

    // Remove the global reference.
    [self release];

     wtFeedbackView = nil;
}

// ========================================================================= //

- (void) doSetText: (NSString*) text;
{
    // Compute by how much the text label frame will change
    // once the text is changed.
    CGSize oldLabelTextSize = self.labelTextSize;
    self.labelTextSize = [text sizeWithFont: self.feedbackLabel.font];

    CGFloat dw = (self.labelTextSize.width  - oldLabelTextSize.width);
    CGFloat dh = (self.labelTextSize.height - oldLabelTextSize.height);

    // Compute the new border view frame.
    CGRect newFrame = self.borderView.frame;

    newFrame.size.width  += dw;
    newFrame.size.height += dh;
    newFrame.origin.x -= dw / 2.0;
    newFrame.origin.y -= dh / 2.0;

    self.textBorderFrame = newFrame;

    // Setting the label text to one or more spaces makes for a
    // smoother text transition, in my experience.
    self.feedbackLabel.text = @" ";

    // Animate the frame changes, then set the label text.
    // NOTE: Using the animationID argument as a holder for the text.

    [UIView beginAnimations: text context: nil];
// [UIView setAnimationDuration: 5.0]; // Animate in slow motion.
    [UIView setAnimationDelegate: self];
    [UIView setAnimationDidStopSelector:
        @selector(textChangeFrameResizingAnimDidStop: finished: context:)];

        self.borderView.frame = self.textBorderFrame;

    [UIView commitAnimations];
}

// ========================================================================= //

- (void) textChangeFrameResizingAnimDidStop: (NSString*) animationID
                                   finished: (NSNumber*) finished
                                    context: (void*) context;
{
    // NOTE: Using the animationID argument as a holder for the text.

    self.feedbackLabel.text = animationID;
    self.borderView.frame = self.textBorderFrame;
}

// ========================================================================= //

// Switches from the progress view to the activity indicator view.

// The default behavior is to do nothing.

// For flexible-style feedback views, animates the transition
// from progress view to activity indicator view, resizing the
// border view as necessary.

// This method is guaranteed to execute in the main thread.

- (void) doSwitchToActivityIndicatorView;
{
    if (self.feedbackKind != WTFeedbackViewKindFlexible ||
        ! self.currentlyShowingProgressView)
    { return; }

    self.currentlyShowingProgressView = NO;

    [UIView beginAnimations: nil context: nil];
// [UIView setAnimationDuration: 5.0]; // Animate in slow motion.
    [UIView setAnimationDelegate: self];
    [UIView setAnimationDidStopSelector:
        @selector(hideProgressAnimDidStop: finished: context:)];

        self.progressView.alpha = 0.0;

    [UIView commitAnimations];
}

// ========================================================================= //

// Switches from the activity indicator view to the progress view.

// The default behavior is to do nothing.

// For flexible-style feedback views, animates the transition
// from activity indicator view to progress view, resizing the
// border view as necessary.

// This method is guaranteed to execute in the main thread.

- (void) doSwitchToProgressView;
{
    if (self.feedbackKind != WTFeedbackViewKindFlexible ||
        self.currentlyShowingProgressView)
    { return; }

    self.currentlyShowingProgressView = YES;

    [UIView beginAnimations: nil context: nil];
// [UIView setAnimationDuration: 5.0]; // Animate in slow motion.
    [UIView setAnimationDelegate: self];
    [UIView setAnimationDidStopSelector:
        @selector(hideActivityAnimDidStop: finished: context:)];

        self.activityIndicatorView.alpha = 0.0;

    [UIView commitAnimations];
}

// ========================================================================= //

// Updates the progress view, if it exists.

// This method is guaranteed to execute in the main thread.

- (void) doUpdateProgress: (NSNumber*) floatObj;
{
    self.progressView.progress = [floatObj floatValue];
}

// ========================================================================= //

- (void) hideProgressAnimDidStop: (NSString*) animationID
                        finished: (NSNumber*) finished
                         context: (void*) context;
{
    self.progressView.alpha = 0.0;

    if ([finished boolValue])
    {
        [UIView beginAnimations: nil context: nil];
// 	[UIView setAnimationDuration: 5.0]; // Uncomment to animate in slow motion
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(toActivityFrameAnimDidStop: finished: context:)];

            self.borderView.frame = self.activityBorderFrame;

        [UIView commitAnimations];
    }
    else
    { [self switchAnimDidStop: nil finished: nil context: NULL]; }

}

// ========================================================================= //

- (void) hideActivityAnimDidStop: (NSString*) animationID
                        finished: (NSNumber*) finished
                         context: (void*) context;
{
     self.activityIndicatorView.alpha = 0.0;
    [self.activityIndicatorView stopAnimating];

    if ([finished boolValue])
    {
        [UIView beginAnimations: nil context: nil];
// 	[UIView setAnimationDuration: 5.0]; // Uncomment to animate in slow motion
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(toProgressFrameAnimDidStop: finished: context:)];

            self.borderView.frame = self.progressBorderFrame;

        [UIView commitAnimations];
    }
    else
    { [self switchAnimDidStop: nil finished: nil context: NULL]; }
}

// ========================================================================= //

- (void) toActivityFrameAnimDidStop: (NSString*) animationID
                           finished: (NSNumber*) finished
                            context: (void*) context;
{
     self.borderView.frame = self.activityBorderFrame;
    [self.activityIndicatorView startAnimating];

    if ([finished boolValue])
    {
        [UIView beginAnimations: nil context: nil];
// 	[UIView setAnimationDuration: 5.0]; // Uncomment to animate in slow motion
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(switchAnimDidStop: finished: context:)];

            self.activityIndicatorView.alpha = 1.0;

        [UIView commitAnimations];
    }
    else
    { [self switchAnimDidStop: nil finished: nil context: NULL]; }
}

// ========================================================================= //

- (void) toProgressFrameAnimDidStop: (NSString*) animationID
                           finished: (NSNumber*) finished
                            context: (void*) context;
{
    self.borderView.frame = self.progressBorderFrame;

    if ([finished boolValue])
    {
        [UIView beginAnimations: nil context: nil];
// 	[UIView setAnimationDuration: 5.0]; // Uncomment to animate in slow motion
        [UIView setAnimationDelegate: self];
        [UIView setAnimationDidStopSelector:
            @selector(switchAnimDidStop: finished: context:)];

            self.progressView.alpha = 1.0;

        [UIView commitAnimations];
    }
    else
    { [self switchAnimDidStop: nil finished: nil context: NULL]; }
}

// ========================================================================= //

- (void) switchAnimDidStop: (NSString*) animationID
                  finished: (NSNumber*) finished
                   context: (void*) context;
{
    if (self.currentlyShowingProgressView)
    {
        self.activityIndicatorView.alpha = 0.0;
        self.borderView.frame = self.progressBorderFrame;
        self.progressView.alpha = 1.0;
    }
    else
    {
        self.progressView.alpha = 0.0;
        self.borderView.frame = self.activityBorderFrame;
        self.activityIndicatorView.alpha = 1.0;
    }
}

// ========================================================================= //


- (void) showAnimDidStop: (NSString*) animationID
                finished: (NSNumber*) finished
                 context: (void*) context;
{
    self.borderView.transform = CGAffineTransformIdentity;
    self.alpha = 1.0;
}

// ========================================================================= //

- (void) removeAnimDidStop: (NSString*) animationID
                  finished: (NSNumber*) finished
                   context: (void*) context;
{
    [[self class] removeView];
}

// ========================================================================= //

#define X_GAP_STYLE_SIMPLE         5.0
#define X_MARGIN_STYLE_SIMPLE     10.0
#define Y_MARGIN_STYLE_SIMPLE     10.0

- (void) layoutSubviewsStyleSimple;
{
    CGSize textSize = self.labelTextSize;

    // 5-pixel horizontal gap between the activity indicator view and the
    // feedback label, indicator to the left, label to the right, both
    // vertically centered in the border. The border view's height is the
    // largest of the two heights (activity indicator and label heights),
    // plus a 10-pixel margin on each side. The border view's width also
    // includes a 10-pixel margin.

    // Calculate the size and position for the border view.
	CGRect borderFrame = CGRectZero;

    borderFrame.size.width = 2.0 * X_MARGIN_STYLE_SIMPLE + X_GAP_STYLE_SIMPLE +
        self.activityIndicatorView.frame.size.width + textSize.width;

    borderFrame.size.height = self.activityIndicatorView.frame.size.height;
    if (textSize.height > borderFrame.size.height)
    { borderFrame.size.height = textSize.height; }

    borderFrame.size.height += (2.0 * Y_MARGIN_STYLE_SIMPLE);

    borderFrame.origin.x =
        floor(0.5 * (self.frame.size.width - borderFrame.size.width));

    borderFrame.origin.y =
        floor(0.5 * (self.frame.size.height - borderFrame.size.height));

    self.borderView.frame = borderFrame;

    // Calculate the position of the indicator.
    CGRect indicatorFrame = self.activityIndicatorView.frame;

	indicatorFrame.origin.x = X_MARGIN_STYLE_SIMPLE;

	indicatorFrame.origin.y =
	    0.5 * (borderFrame.size.height - indicatorFrame.size.height);

    self.activityIndicatorView.frame = indicatorFrame;

    // Calculate the position of the label.

    CGRect labelFrame = self.feedbackLabel.frame;

    labelFrame.size = textSize;

    labelFrame.origin.x = indicatorFrame.origin.x + X_GAP_STYLE_SIMPLE +
        indicatorFrame.size.width;

    labelFrame.origin.y =
        floor(0.5 * (borderFrame.size.height - labelFrame.size.height));

    self.feedbackLabel.frame = labelFrame;
}

// ========================================================================= //

#define Y_MARGIN                   15.0

- (void) layoutCenteredLabel;
{
    // Calculate the position of the label.

    CGRect labelFrame = self.feedbackLabel.frame;

    labelFrame.size = self.labelTextSize;

    labelFrame.origin.x =
        floor(0.5 * (self.borderView.frame.size.width - labelFrame.size.width));

    labelFrame.origin.y = self.borderView.frame.size.height - Y_MARGIN -
        labelFrame.size.height;

    self.feedbackLabel.frame = labelFrame;
}

// ========================================================================= //

#define Y_GAP_KIND_ACTIVITY        10.0
#define X_MARGIN_KIND_ACTIVITY     20.0

- (void) layoutBorderAndActivity;
{
    CGSize textSize = self.labelTextSize;

    // 10-pixel vertical gap between the activity indicator view and the
    // feedback label, both horizontally centered in the border. The
    // combination of the two is centered both horizontally and vertically
    // within the border view, with a margin all around.

    // Calculate the size and position for the border view.
	CGRect borderFrame = CGRectZero;

    borderFrame.size.width = textSize.width;
    if (self.activityIndicatorView.frame.size.width > borderFrame.size.width)
    { borderFrame.size.width = self.activityIndicatorView.frame.size.width; }

    borderFrame.size.width += (2.0 * X_MARGIN_KIND_ACTIVITY);

    borderFrame.size.height = self.activityIndicatorView.frame.size.height +
        Y_GAP_KIND_ACTIVITY + textSize.height + 2.0 * Y_MARGIN;

    borderFrame.origin.x =
        floor(0.5 * (self.frame.size.width - borderFrame.size.width));

    borderFrame.origin.y =
        floor(0.5 * (self.frame.size.height - borderFrame.size.height));

    self.borderView.frame = borderFrame;

    // Calculate the position of the indicator.
    CGRect indicatorFrame = self.activityIndicatorView.frame;

	indicatorFrame.origin.x =
	    0.5 * (self.borderView.frame.size.width - indicatorFrame.size.width);

	indicatorFrame.origin.y = Y_MARGIN;

    self.activityIndicatorView.frame = indicatorFrame;
    self.activityBorderFrame = self.borderView.frame;
}

// ========================================================================= //

#define Y_GAP_KIND_PROGRESS        10.0
#define X_MARGIN_KIND_PROGRESS     30.0
#define Y_MARGIN_KIND_PROGRESS     15.0

- (void) layoutBorderAndProgress;
{
    CGSize textSize = self.labelTextSize;

    // 10-pixel vertical gap between the progress view and the
    // feedback label, both horizontally centered in the border. The
    // combination of the two is centered both horizontally and vertically
    // within the border view, with a margin all around.

    // Calculate the size and position for the border view.
	CGRect borderFrame = CGRectZero;

    borderFrame.size.width = textSize.width + 2.0 * X_MARGIN_KIND_PROGRESS;

    borderFrame.size.height = self.progressView.frame.size.height +
        Y_GAP_KIND_PROGRESS + textSize.height + 2.0 * Y_MARGIN_KIND_PROGRESS;

    borderFrame.origin.x =
        floor(0.5 * (self.frame.size.width - borderFrame.size.width));

    borderFrame.origin.y =
        floor(0.5 * (self.frame.size.height - borderFrame.size.height));

    self.borderView.frame = borderFrame;

    // Calculate the position of the progress view.
    CGRect progressFrame = self.progressView.frame;

    progressFrame.size.width =
        self.borderView.frame.size.width - X_MARGIN_KIND_PROGRESS;

	progressFrame.origin.x = X_MARGIN_KIND_PROGRESS / 2.0;
	progressFrame.origin.y = Y_MARGIN_KIND_PROGRESS;

    self.progressView.frame = progressFrame;
    self.progressBorderFrame = self.borderView.frame;
}

// ========================================================================= //

@end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //





// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //
                                #pragma mark -
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //

@implementation UIApplication (KeyboardView)

//  Copyright Matt Gallagher 2009. All rights reserved.

//  Permission is given to use this source code file, free of charge, in any
//  project, commercial or otherwise, entirely at your risk, with the condition
//  that any redistribution (in part or whole) of source code must retain
//  this copyright and permission notice. Attribution in compiled projects is
//  appreciated but not required.

- (UIView*) keyboardView;
{
	NSArray* windows = [self windows];
	for (UIWindow* window in [windows reverseObjectEnumerator])
	{
		for (UIView* view in [window subviews])
		{
			if (! strcmp(object_getClassName(view), "UIKeyboard"))
			{
				return view;
			}
		}
	}

	return nil;
}

@end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ //
